package org.commonjava.tensor.agg;

import static org.apache.commons.lang.StringUtils.join;

import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;

import javax.enterprise.context.ApplicationScoped;
import javax.inject.Inject;

import org.apache.maven.graph.common.ref.ArtifactRef;
import org.apache.maven.graph.common.ref.ProjectRef;
import org.apache.maven.graph.common.ref.ProjectVersionRef;
import org.apache.maven.graph.common.version.SingleVersion;
import org.apache.maven.graph.effective.EProjectCycle;
import org.apache.maven.graph.effective.EProjectCycle.Builder;
import org.apache.maven.graph.effective.EProjectGraph;
import org.apache.maven.graph.effective.EProjectNet;
import org.apache.maven.graph.effective.EProjectWeb;
import org.apache.maven.graph.effective.rel.DependencyRelationship;
import org.apache.maven.graph.effective.rel.ProjectRelationship;
import org.apache.maven.graph.spi.GraphDriverException;
import org.commonjava.cdi.util.weft.ExecutorConfig;
import org.commonjava.tensor.data.TensorDataException;
import org.commonjava.tensor.data.TensorDataManager;
import org.commonjava.tensor.discover.ProjectRelationshipDiscoverer;
import org.commonjava.tensor.event.TensorEventFunnel;
import org.commonjava.util.logging.Logger;

@ApplicationScoped
public class DefaultGraphAggregator
    implements GraphAggregator
{

    private final Logger logger = new Logger( getClass() );

    @Inject
    private TensorDataManager dataManager;

    @Inject
    private ProjectRelationshipDiscoverer discoverer;

    @Inject
    private TensorEventFunnel funnel;

    @Inject
    @ExecutorConfig( named = "tensor-main", threads = 4 )
    private ExecutorService executor;

    protected DefaultGraphAggregator()
    {
    }

    public DefaultGraphAggregator( final TensorDataManager dataManager, final ProjectRelationshipDiscoverer discoverer,
                                   final TensorEventFunnel funnel )
    {
        this.dataManager = dataManager;
        this.discoverer = discoverer;
        this.funnel = funnel;
    }

    /* (non-Javadoc)
     * @see org.commonjava.tensor.agg.GraphCrawler#connectIncompleteSubgraphs(org.apache.maven.graph.effective.EProjectGraph, org.apache.maven.graph.effective.filter.ProjectRelationshipFilter, org.commonjava.tensor.agg.DiscoveryConfig, boolean)
     */
    @Override
    public EProjectGraph connectSubgraphs( final EProjectGraph graph, final AggregationOptions config )
        throws TensorDataException
    {
        return connectAndFilter( graph, config, graph.getRoot() );
    }

    @Override
    public EProjectWeb connectSubwebs( final EProjectWeb web, final AggregationOptions config,
                                       final ProjectVersionRef... roots )
        throws TensorDataException
    {
        return connectAndFilter( web, config, roots );
    }

    @Override
    public Map<ProjectRef, ProjectRefCollection> collectProjectReferences( final AggregationOptions options,
                                                                           final AggregatorConfig config )
        throws TensorDataException
    {
        final ProjectVersionRef[] refs = config.getRoots();
        EProjectWeb web = dataManager.getProjectWeb( refs );
        if ( web != null )
        {
            //                if ( options.isDiscoveryEnabled() )
            //                {
            web = connectSubwebs( web, options, refs );
            //                }

        }

        final Map<ProjectRef, ProjectRefCollection> projects = new HashMap<ProjectRef, ProjectRefCollection>();

        final Set<ProjectRelationship<?>> rels = web.getAllRelationships();
        for ( final ProjectRelationship<?> rel : rels )
        {
            if ( !( rel instanceof DependencyRelationship ) )
            {
                continue;
            }

            final ProjectVersionRef orig = rel.getDeclaring();
            final ProjectVersionRef pvr = orig.asProjectVersionRef();
            final ProjectRef r = rel.getDeclaring()
                                    .asProjectRef();
            ProjectRefCollection prc = projects.get( r );
            if ( prc == null )
            {
                prc = new ProjectRefCollection();
                projects.put( r, prc );
            }

            prc.addVersionRef( pvr );

            final ArtifactRef tar = rel.getTargetArtifact()
                                       .setOptional( false );

            final ProjectRef tr = tar.asProjectRef();
            ProjectRefCollection tprc = projects.get( tr );
            if ( tprc == null )
            {
                tprc = new ProjectRefCollection();
                projects.put( tr, tprc );
            }

            tprc.addArtifactRef( tar );
        }

        return projects;
    }

    private <T extends EProjectNet> T connectAndFilter( final T net, final AggregationOptions config,
                                                        final ProjectVersionRef... roots )
        throws TensorDataException
    {
        if ( net != null )
        {
            if ( config.isDiscoveryEnabled() )
            {
                final LinkedList<EProjectNet> nets = new LinkedList<EProjectNet>();
                nets.add( net );

                final Set<ProjectVersionRef> missing = new HashSet<ProjectVersionRef>();

                int changeCount = 0;
                do
                {
                    changeCount =
                        connectSubs( net, nets, new HashSet<EProjectCycle.Builder>(), config,
                                     new HashSet<ProjectVersionRef>(), missing );

                    logger.info( "%d changes...re-running to see if we can discover a bit more.", changeCount );
                }
                while ( changeCount > 0 );
            }

            if ( config.getFilter() != null )
            {
                try
                {
                    return net.filteredInstance( config.getFilter() );
                }
                catch ( final GraphDriverException e )
                {
                    throw new TensorDataException( "Failed to construct filtered web for roots: %s. Reason: %s", e,
                                                   join( roots, ", " ), e.getMessage() );
                }
            }
            else
            {
                return net;
            }
        }

        return net;
    }

    private int connectSubs( final EProjectNet graph, final LinkedList<EProjectNet> graphs,
                             final Set<EProjectCycle.Builder> cycleBuilders, final AggregationOptions config,
                             final Set<ProjectVersionRef> cycleParticipants, final Set<ProjectVersionRef> missing )
        throws TensorDataException
    {
        int changed = 0;
        //        logger.info( "Reconnecting: %s using filter: %s", graph.getRoot(), filter );
        final Set<ProjectVersionRef> incomplete = new HashSet<ProjectVersionRef>( graph.getIncompleteSubgraphs() );
        if ( !incomplete.isEmpty() )
        {
            changed += discover( incomplete, graph, graphs, cycleBuilders, config, cycleParticipants, missing );
        }

        final Set<ProjectVersionRef> variables = new HashSet<ProjectVersionRef>( graph.getVariableSubgraphs() );
        if ( !variables.isEmpty() )
        {
            final Set<ProjectVersionRef> todo = new HashSet<ProjectVersionRef>();
            for ( final ProjectVersionRef variable : variables )
            {
                final ProjectVersionRef ref = discoverer.resolveSpecificVersion( variable );
                if ( ref.getVersionSpec()
                        .isConcrete() )
                {
                    try
                    {
                        graph.selectVersionFor( variable, (SingleVersion) ref.getVersionSpec() );
                    }
                    catch ( final GraphDriverException e )
                    {
                        throw new TensorDataException(
                                                       "Failed to select %s for variable project reference: %s. Reason: %s",
                                                       e, ref, variable );
                    }

                    if ( graph.getDriver()
                              .isMissing( ref ) )
                    {
                        todo.add( ref );
                    }
                }
            }

            if ( !todo.isEmpty() )
            {
                changed += discover( todo, graph, graphs, cycleBuilders, config, cycleParticipants, missing );
            }
        }

        return changed;
    }

    private int discover( final Set<ProjectVersionRef> toDiscover, final EProjectNet graph,
                          final LinkedList<EProjectNet> graphs, final Set<Builder> cycleBuilders,
                          final AggregationOptions config, final Set<ProjectVersionRef> cycleParticipants,
                          final Set<ProjectVersionRef> missing )
        throws TensorDataException
    {
        Set<ProjectVersionRef> incomplete = toDiscover;
        for ( final Iterator<ProjectVersionRef> it = incomplete.iterator(); it.hasNext(); )
        {
            final ProjectVersionRef ref = it.next();
            if ( missing.contains( ref ) )
            {
                it.remove();
            }
        }

        incomplete = resolveVariableVersions( incomplete );

        incomplete =
            dataManager.pathFilter( config.getFilter(), incomplete, graph.getDriver()
                                                                         .getRoots()
                                                                         .toArray( new ProjectVersionRef[] {} ) );

        logger.info( "Performing discovery and cycle-detection on %d missing subgraphs: %s", incomplete.size(),
                     join( incomplete, ", " ) );

        final Set<DiscoveryRunnable> runnables = new HashSet<DiscoveryRunnable>( incomplete.size() );
        final CountDownLatch latch = new CountDownLatch( incomplete.size() );

        final Set<ProjectVersionRef> roMissing = Collections.unmodifiableSet( missing );
        for ( final ProjectVersionRef ref : incomplete )
        {
            final DiscoveryRunnable runnable =
                new DiscoveryRunnable( ref, graph, graphs, cycleBuilders, config, cycleParticipants, roMissing,
                                       discoverer, funnel, latch );

            executor.execute( runnable );
            runnables.add( runnable );
        }

        try
        {
            latch.await();
        }
        catch ( final InterruptedException e )
        {
            logger.error( "Interrupted on subgraph reconnect for: %s", graph );
        }

        int changeCount = 0;
        for ( final DiscoveryRunnable r : runnables )
        {
            missing.addAll( r.getNewMissing() );
            if ( r.isChanged() )
            {
                changeCount++;
            }
        }

        return changeCount;
    }

    private Set<ProjectVersionRef> resolveVariableVersions( final Set<ProjectVersionRef> incomplete )
    {
        final Set<ProjectVersionRef> result = new HashSet<ProjectVersionRef>();
        for ( final ProjectVersionRef ref : incomplete )
        {
            // NOTE: Side effect of this verifies that the version parses correctly.
            try
            {
                final ProjectVersionRef resolved = discoverer.resolveSpecificVersion( ref );
                result.add( resolved );
            }
            catch ( final TensorDataException e )
            {
                logger.error( "Failed to resolve specific version for: %s. Reason: %s", e, ref, e.getMessage() );
                result.add( ref );
            }
        }

        return result;
    }

}
