package org.commonjava.tensor.agg;

import java.util.HashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.concurrent.CountDownLatch;

import org.apache.maven.graph.common.ref.ProjectVersionRef;
import org.apache.maven.graph.common.version.InvalidVersionSpecificationException;
import org.apache.maven.graph.effective.EProjectCycle;
import org.apache.maven.graph.effective.EProjectCycle.Builder;
import org.apache.maven.graph.effective.EProjectNet;
import org.apache.maven.graph.effective.filter.ProjectRelationshipFilter;
import org.apache.maven.graph.effective.ref.EProjectKey;
import org.apache.maven.graph.effective.rel.ProjectRelationship;
import org.apache.maven.graph.spi.GraphDriverException;
import org.commonjava.tensor.data.TensorDataException;
import org.commonjava.tensor.discover.DiscoveryConfig;
import org.commonjava.tensor.discover.ProjectRelationshipDiscoverer;
import org.commonjava.tensor.event.TensorEventFunnel;
import org.commonjava.util.logging.Logger;

public class DiscoveryRunnable
    implements Runnable
{

    private final ProjectVersionRef originalRef;

    private final EProjectNet graph;

    private final LinkedList<EProjectNet> graphs;

    private final AggregationOptions config;

    private final CountDownLatch latch;

    private final Logger logger = new Logger( getClass() );

    private final Set<ProjectVersionRef> roMissing;

    private final Set<ProjectVersionRef> newMissing = new HashSet<ProjectVersionRef>();

    private final Set<ProjectVersionRef> cycleParticipants;

    private final Set<Builder> cycleBuilders;

    private boolean changed = false;

    private final ProjectRelationshipDiscoverer discoverer;

    private final TensorEventFunnel funnel;

    public DiscoveryRunnable( final ProjectVersionRef ref, final EProjectNet graph,
                              final LinkedList<EProjectNet> graphs, final Set<EProjectCycle.Builder> cycleBuilders,
                              final AggregationOptions config, final Set<ProjectVersionRef> cycleParticipants,
                              final Set<ProjectVersionRef> missing, final ProjectRelationshipDiscoverer discoverer,
                              final TensorEventFunnel funnel, final CountDownLatch latch )
    {
        this.originalRef = ref;
        this.graph = graph;
        this.graphs = graphs;
        this.cycleBuilders = cycleBuilders;
        this.config = config;
        this.cycleParticipants = cycleParticipants;
        this.roMissing = missing;
        this.discoverer = discoverer;
        this.funnel = funnel;
        this.latch = latch;
    }

    public boolean isChanged()
    {
        return changed;
    }

    @Override
    public void run()
    {
        final ProjectRelationshipFilter filter = config.getFilter();
        final ProjectVersionRef ref = originalRef;

        logger.info( "\n\n\n\nProcessing missing project: %s\n\n\n\n", ref );

        try
        {
            // NOTE: This is now handled in the DefaultGraphAggregator before discover runnables are created.

            //            if ( filter != null )
            //            {
            //                logger.info( "Looking for paths to missing project: %s in network: %s filtered by: %s", ref, graph,
            //                             filter );
            //                final Set<List<ProjectRelationship<?>>> paths = graph.getPathsTo( ref );
            //                if ( paths != null )
            //                {
            //                    boolean pass = false;
            //                    nextPath: for ( final List<ProjectRelationship<?>> path : paths )
            //                    {
            //                        ProjectRelationshipFilter f = filter;
            //                        for ( final ProjectRelationship<?> rel : path )
            //                        {
            //                            if ( !f.accept( rel ) )
            //                            {
            //                                logger.info( "Path: %s rejected, trying any others that remain.", path );
            //                                continue nextPath;
            //                            }
            //
            //                            f = f.getChildFilter( rel );
            //                        }
            //
            //                        pass = true;
            //                        break;
            //                    }
            //
            //                    if ( !pass )
            //                    {
            //                        logger.info( "No path to project: %s was found in network: %s filtered by: %s", ref, graph,
            //                                     filter );
            //                        return;
            //                    }
            //                }
            //            }

            //            logger.info( "%s marked as missing in: %s", ref, graph.getRoot() );

            final long timeoutMillis = config.getDiscoveryTimeoutMillis();
            final DiscoveryConfig discoveryConfig = config.getDiscoveryConfig();

            final Set<EProjectCycle> cycles = graph.getCycles();
            if ( cycles != null )
            {
                for ( final EProjectCycle cycle : cycles )
                {
                    final Set<ProjectVersionRef> participants = cycle.getAllParticipatingProjects();
                    cycleParticipants.addAll( participants );
                }
            }

            if ( cycleParticipants.contains( ref ) )
            {
                logger.warn( "Dependency-cycle participant detected: %s. It was marked incomplete in network: %s. Skipping reconnection.",
                             ref, graph );

                // not TECHNICALLY correct, but it renders the graph VERY difficult to use, so...
                markMissing( ref );
                return;
            }

            final Set<ProjectRelationship<?>> rels = graph.getRelationshipsTargeting( originalRef );
            if ( detectCycles( ref, rels, graphs, cycleBuilders, filter, cycleParticipants ) )
            {
                return;
            }

            if ( discoverer != null && !roMissing.contains( ref ) )
            {
                // kick off the async process of resolving metadata and building the graph.
                // then, wait for the event funnel to capture something telling us to re-check
                // for the project graph.
                discoverer.discoverRelationships( ref, discoveryConfig );
                funnel.waitForGraph( ref, timeoutMillis );

                logger.info( "Connecting: %s to: %s", graph, ref );
                if ( !graph.connectFor( new EProjectKey( ref ) ) )
                {
                    markMissing( ref );
                }
                else
                {
                    changed = true;
                }
            }
        }
        catch ( final InvalidVersionSpecificationException e )
        {
            logger.error( "Cannot discover subgraph for: %s. Reason: %s.", e, ref, e.getMessage() );
        }
        catch ( final GraphDriverException e )
        {
            logger.error( "Failed to connect graph for: %s to our network: %s. Reason: %s", e, ref, graph,
                          e.getMessage() );
        }
        catch ( final TensorDataException e )
        {
            logger.error( "Failed to connect graph for: %s to our network: %s. Reason: %s", e, ref, graph,
                          e.getMessage() );
        }
        finally
        {
            latch.countDown();
        }
    }

    private void markMissing( final ProjectVersionRef ref )
    {
        newMissing.add( ref );
        if ( !originalRef.equals( ref ) )
        {
            newMissing.add( originalRef );
        }
    }

    private boolean detectCycles( final ProjectVersionRef ref, final Set<ProjectRelationship<?>> rels,
                                  final List<EProjectNet> graphs, final Set<Builder> cycleBuilders,
                                  final ProjectRelationshipFilter filters,
                                  final Set<ProjectVersionRef> cycleParticipants )
    {
        boolean foundCycle = false;

        for ( final EProjectCycle.Builder cycleBuilder : cycleBuilders )
        {
            final int idx = cycleBuilder.indexOf( ref );
            if ( idx > -1 )
            {
                // CYCLE!!!
                final EProjectCycle.Builder cb = new EProjectCycle.Builder( cycleBuilder, idx );
                for ( final ProjectRelationship<?> rel : rels )
                {
                    cycleParticipants.add( rel.getDeclaring() );
                    cycleParticipants.add( rel.getTarget()
                                              .asProjectVersionRef() );

                    if ( !filters.accept( rel ) )
                    {
                        continue;
                    }

                    cb.with( rel );

                    final EProjectCycle cycle = cb.build();
                    //                    logger.info( "CYCLE! %s", cycle );

                    for ( final EProjectNet g : graphs )
                    {
                        g.addCycle( cycle );
                    }

                    cb.withoutLast();
                    foundCycle = true;
                }
            }
        }

        return foundCycle;
    }

    public Set<ProjectVersionRef> getNewMissing()
    {
        return newMissing;
    }

}
